# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"Array": "00_typing.ipynb",
         "Int": "00_typing.ipynb",
         "Float": "00_typing.ipynb",
         "ComplexFloat": "00_typing.ipynb",
         "Settings": "00_typing.ipynb",
         "SDict": "00_typing.ipynb",
         "SCoo": "00_typing.ipynb",
         "SDense": "00_typing.ipynb",
         "SType": "00_typing.ipynb",
         "Model": "00_typing.ipynb",
         "ModelFactory": "00_typing.ipynb",
         "Models": "00_typing.ipynb",
         "is_float": "00_typing.ipynb",
         "is_complex": "00_typing.ipynb",
         "is_complex_float": "00_typing.ipynb",
         "is_sdict": "00_typing.ipynb",
         "is_scoo": "00_typing.ipynb",
         "is_sdense": "00_typing.ipynb",
         "is_model": "00_typing.ipynb",
         "is_model_factory": "00_typing.ipynb",
         "validate_model": "00_typing.ipynb",
         "is_stype": "00_typing.ipynb",
         "is_singlemode": "00_typing.ipynb",
         "is_multimode": "00_typing.ipynb",
         "is_mixedmode": "00_typing.ipynb",
         "sdict": "00_typing.ipynb",
         "scoo": "00_typing.ipynb",
         "sdense": "00_typing.ipynb",
         "modelfactory": "00_typing.ipynb",
         "block_diag": "02_utils.ipynb",
         "clean_string": "02_utils.ipynb",
         "copy_settings": "02_utils.ipynb",
         "validate_settings": "02_utils.ipynb",
         "try_float": "02_utils.ipynb",
         "flatten_dict": "02_utils.ipynb",
         "unflatten_dict": "02_utils.ipynb",
         "get_ports": "02_utils.ipynb",
         "get_port_combinations": "02_utils.ipynb",
         "get_settings": "02_utils.ipynb",
         "grouped_interp": "02_utils.ipynb",
         "merge_dicts": "02_utils.ipynb",
         "mode_combinations": "02_utils.ipynb",
         "reciprocal": "02_utils.ipynb",
         "rename_params": "02_utils.ipynb",
         "rename_ports": "02_utils.ipynb",
         "update_settings": "02_utils.ipynb",
         "validate_not_mixedmode": "02_utils.ipynb",
         "validate_multimode": "02_utils.ipynb",
         "validate_sdict": "02_utils.ipynb",
         "get_inputs_outputs": "02_utils.ipynb",
         "hash_dict": "02_utils.ipynb",
         "multimode": "04_multimode.ipynb",
         "singlemode": "04_multimode.ipynb",
         "straight": "05_models.ipynb",
         "coupler": "05_models.ipynb",
         "unitary": "05_models.ipynb",
         "passthru": "05_models.ipynb",
         "copier": "05_models.ipynb",
         "get_models": "05_models.ipynb",
         "models": "05_models.ipynb",
         "BaseModel": "06a_netlist.ipynb",
         "Component": "06a_netlist.ipynb",
         "PortEnum": "06a_netlist.ipynb",
         "Placement": "06a_netlist.ipynb",
         "Route": "06a_netlist.ipynb",
         "Netlist": "06a_netlist.ipynb",
         "RecursiveNetlist": "06a_netlist.ipynb",
         "netlist": "06a_netlist.ipynb",
         "load_netlist": "06a_netlist.ipynb",
         "load_recursive_netlist": "06a_netlist.ipynb",
         "get_netlist_instances_by_prefix": "06a_netlist.ipynb",
         "get_component_instances": "06a_netlist.ipynb",
         "remove_unused_instances": "06b_netlist_cleaning.ipynb",
         "create_dag": "07_circuit.ipynb",
         "draw_dag": "07_circuit.ipynb",
         "find_root": "07_circuit.ipynb",
         "find_leaves": "07_circuit.ipynb",
         "circuit": "07_circuit.ipynb",
         "NetlistDict": "07_circuit.ipynb",
         "CircuitInfo": "07_circuit.ipynb",
         "RecursiveNetlistDict": "07_circuit.ipynb",
         "get_required_circuit_models": "07_circuit.ipynb",
         "circuit_backends": "08_backends.ipynb",
         "evaluate_circuit": "08a_backends_default.ipynb",
         "solve_klu": "08b_backends_klu.ipynb",
         "mul_coo": "08b_backends_klu.ipynb",
         "evaluate_circuit_klu": "08b_backends_klu.ipynb",
         "split_port": "08c_backends_additive.ipynb",
         "graph_edges": "08c_backends_additive.ipynb",
         "prune_internal_output_nodes": "08c_backends_additive.ipynb",
         "get_possible_paths": "08c_backends_additive.ipynb",
         "path_lengths": "08c_backends_additive.ipynb",
         "evaluate_circuit_additive": "08c_backends_additive.ipynb",
         "mse": "09a_nn_loss.ipynb",
         "huber_loss": "09a_nn_loss.ipynb",
         "l2_reg": "09a_nn_loss.ipynb",
         "cartesian_product": "09b_nn_utils.ipynb",
         "denormalize": "09b_nn_utils.ipynb",
         "norm": "09b_nn_utils.ipynb",
         "get_normalization": "09b_nn_utils.ipynb",
         "get_df_columns": "09b_nn_utils.ipynb",
         "normalize": "09b_nn_utils.ipynb",
         "preprocess": "09c_nn_core.ipynb",
         "dense": "09c_nn_core.ipynb",
         "generate_dense_weights": "09c_nn_core.ipynb",
         "load_nn_weights_json": "09d_nn_io.ipynb",
         "save_nn_weights_json": "09d_nn_io.ipynb",
         "get_available_sizes": "09d_nn_io.ipynb",
         "get_dense_weights_path": "09d_nn_io.ipynb",
         "get_norm_path": "09d_nn_io.ipynb",
         "load_nn_dense": "09d_nn_io.ipynb",
         "MAGIC_COMMENTS": "98_make_docs.ipynb",
         "load_nb": "98_make_docs.ipynb",
         "repository_path": "98_make_docs.ipynb",
         "docs_path": "98_make_docs.ipynb",
         "save_nb": "98_make_docs.ipynb",
         "strip_metadata": "98_make_docs.ipynb",
         "iter_code_cells": "98_make_docs.ipynb",
         "first_code_cell": "98_make_docs.ipynb",
         "get_default_exp": "98_make_docs.ipynb",
         "iter_function_names": "98_make_docs.ipynb",
         "docs_copy_nb": "98_make_docs.ipynb",
         "list_notebooks": "98_make_docs.ipynb",
         "list_zips": "98_make_docs.ipynb",
         "docs_copy_dir": "98_make_docs.ipynb",
         "docs_copy_simulations": "98_make_docs.ipynb",
         "get_toc_part": "98_make_docs.ipynb",
         "make_docs": "98_make_docs.ipynb",
         "__author__": "99_init.ipynb",
         "__version__": "99_init.ipynb"}

modules = ["typing_.py",
           "patched.py",
           "utils.py",
           "multimode.py",
           "models.py",
           "netlist.py",
           "netlist_cleaning.py",
           "circuit.py",
           "backends/__init__.py",
           "backends/default.py",
           "backends/klu.py",
           "backends/additive.py",
           "nn/__init__.py",
           "nn/loss.py",
           "nn/utils.py",
           "nn/core.py",
           "nn/io.py",
           "make_docs.py",
           "__init__.py"]

doc_url = "https://flaport.github.io/sax/"

git_url = "https://github.com/flaport/sax/tree/master/"

def custom_doc_links(name): return None
